// =======================================================================
//
// MAS22_Apocalypse.cpp
// Template for MATT
// CBP 03-08-22
// Following code from MLP completed by CBP 
// 11-08-22
// 
// =======================================================================

#include "MAS22_Apocalypse.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"
#include "Components/CapsuleComponent.h"
#include "Components/StaticMeshComponent.h"
//#include "DestructibleComponent.h"


// defines for indexing into modelParameters array -----------------------
#define INDEX_MASS 0
#define INDEX_GRAVY 1
#define INDEX_LVELOCITY 2
#define INDEX_RHEIGHT 3
#define INDEX_BHEIGHT 4
#define INDEX_BFORCE 5
#define INDEX_RTIME 6

#define STARTING 5
#define RISING 0
#define PAUSING 1
#define FALLING 2
#define BRAKING 3
#define STOPPED 4

// defines index into modelInitialConditions array -------------------------
#define INDEX_INITDISPZ 0
#define INDEX_INITVELYZ 1

// Number of equations in rhs ----------------------------------------------
#define NREQUNS 2

// -------------------------------------------------------------------------

AMAS22_Apocalypse::AMAS22_Apocalypse() {

	PrimaryActorTick.bCanEverTick = false;

	// Main collider
	CapsuleCollider = CreateDefaultSubobject<UCapsuleComponent>("CapsuleCollider");
	CapsuleCollider->SetupAttachment(GetRootComponent());
	CapsuleCollider->SetCapsuleHalfHeight(2000);
	CapsuleCollider->SetCapsuleRadius(115);

	// Create and attach the meshes

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("StaticMesh'/Game/Apocalypse/apocTowerV2.apocTowerV2'"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	Tower = CreateDefaultSubobject<UStaticMeshComponent>("Tower");
	Tower->SetStaticMesh(Asset1);
	Tower->SetupAttachment(CapsuleCollider);
	Tower->SetRelativeScale3D(FVector(2, 2, 41.5));

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset2(TEXT("StaticMesh'/Game/Apocalypse/apocTowerV2Top.apocTowerV2Top'"));
	UStaticMesh* Asset2 = MeshAsset2.Object;

	TowerTop = CreateDefaultSubobject<UStaticMeshComponent>("TowerTop");
	TowerTop->SetStaticMesh(Asset2);
	TowerTop->SetupAttachment(CapsuleCollider);
	TowerTop->SetRelativeLocation(FVector(0, 0, 5120));
	TowerTop->SetRelativeScale3D(FVector(3, 3, 2));

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset3(TEXT("StaticMesh'/Game/Apocalypse/apocCarV2.apocCarV2'"));
	UStaticMesh* Asset3 = MeshAsset3.Object;

	Car = CreateDefaultSubobject<UStaticMeshComponent>("Car");
	Car->SetStaticMesh(Asset3);
	Car->SetupAttachment(CapsuleCollider);
	Car->SetRelativeLocation(FVector(155, 0, 0));
	Car->SetRelativeScale3D(FVector(2));

	//	DestroyedCarMesh = CreateDefaultSubobject<UDestructibleComponent>("DestroyedCarMesh");
	//	DestroyedCarMesh->SetupAttachment(CarMesh);
	//	DestroyedCarMesh->SetRelativeScale3D(FVector(2));
	//	DestroyedCarMesh->SetVisibility(false);
	//	DestroyedCarMesh->SetSimulatePhysics(false);

	CarStartLocation = Car->GetRelativeLocation();

}

void AMAS22_Apocalypse::ActorInitialize() { // ---------------------------
	Super::ActorInitialize();

	// Call to function to apply the config parameters -------------------
	ApplyConfig();

	// Apply global logging ----------------------------------------------
	InitGlobalLog();
	flog("Hi from AMAS22_Apocalypse\n");

	u = new double[NREQUNS];
	dudt = new double[NREQUNS];
	yout = new double[NREQUNS]; // Required for Runge-Kutta

	SetModelICs();
	UpdateModelICs();

	GetDefaultParameters();
	SetModelParameters();
	UpdateModelParameters();

	InitializeSolver();

	// Create the solver object and the right-hand-side Functor 0---------
	solverRK = new MAS22_Solver_RK();
	solverRKF = new MAS22_Solver_RKF();

	rightHandSide = new rhs_Apocalypse(_mass);
	// In above fill out param list. Precede each by underscore

	state = STARTING;

}

void AMAS22_Apocalypse::SetModelParameters() { //-------------------------
	modelParametersStruct line;
	modelParameters.Empty();

	line.val = mass;
	line.name = "mass";
	modelParameters.Add(line);

	line.val = gravity;
	line.name = "gravity";
	modelParameters.Add(line);

	line.val = liftVelocity;
	line.name = "liftVelocity";
	modelParameters.Add(line);

	line.val = releaseHeight;
	line.name = "releaseHeight";
	modelParameters.Add(line);

	line.val = brakingHeight;
	line.name = "breakingHeight";
	modelParameters.Add(line);

	line.val = brakingForce;
	line.name = "breakingForce";
	modelParameters.Add(line);

	line.val = releaseTime;
	line.name = "releaseTime";
	modelParameters.Add(line);

}

void AMAS22_Apocalypse::UpdateModelParameters() { // ---------------------
	// change suffix P to prefix _
	_mass = modelParameters[INDEX_MASS].val;
	_gravity = modelParameters[INDEX_GRAVY].val;
	_liftVelocity = modelParameters[INDEX_LVELOCITY].val;
	_releaseHeight = modelParameters[INDEX_RHEIGHT].val;
	_brakingHeight = modelParameters[INDEX_BHEIGHT].val;
	_brakingForce = modelParameters[INDEX_BFORCE].val;
	_releaseTime = modelParameters[INDEX_RTIME].val;
	// Add lines for all parameters

	delete rightHandSide;
	rightHandSide = new rhs_Apocalypse(_mass);
	// Copy and paste from line in ActorInitialize();

}
void AMAS22_Apocalypse::GetDefaultParameters() { // ----------------------
	_mass = mass;
	_gravity = gravity;
	_liftVelocity = liftVelocity;
	_releaseHeight = releaseHeight;
	_brakingHeight = brakingHeight;
	_brakingForce = brakingForce;
	_releaseTime = releaseTime;
	// Add lines for all parameters

}

// Model initial conditions ----------------------------------------------

void AMAS22_Apocalypse::SetModelICs() {
	Super::SetModelICs();

	modelParametersStruct icLine;
	modelInitialConditions.Empty();			// allows for hard reset

	icLine.val = initDispZ;
	icLine.name = "dispz";
	modelInitialConditions.Add(icLine);

	icLine.val = initVelyZ;
	icLine.name = "velyz";
	modelInitialConditions.Add(icLine);

}


void AMAS22_Apocalypse::UpdateModelICs() { // ----------------------------
	initDispZ = modelInitialConditions[INDEX_INITDISPZ].val;
	initVelyZ = modelInitialConditions[INDEX_INITVELYZ].val;
	// Add lines for all initial conditions

	// last line must be
	InitializeSolver(); // to set the ICs
}


void AMAS22_Apocalypse::InitializeSolver() { // --------------------------

	flag = 2;
	t = 0.0;
	time = 0;
	// Call to orig rkf code <<<<<<<<<<<<<<<<<<<<<<<<<<<<
	relerr = 10 * sqrt(solverRKF->r8_epsilon());
	abserr = 10 * sqrt(solverRKF->r8_epsilon());

	printf("AMAS22_Apocalypse::InitializeSolver() relerr = %e, abserr  = %e", relerr, abserr);

	// Set the physically determined constants ------------------------------
	// Apply your initial conditions to the state vector u[...]
	u[0] = initDispZ;
	u[1] = initVelyZ;


	// Grab the state values and copy into your system variabkes
	// (needed by model reset)
	dispZ = u[0];
	velyZ = u[1];

}


void AMAS22_Apocalypse::BuildHUDLines() { // -----------------------------

	HUDLine line;

	HUDLines.Empty();
	bHUD_exists = true;
	if (bHUD_exists) {
		Super::BuildHUDLines();
	}

	// Add more lines specific to the model -----------------------------------

	line.message = FString::Printf(TEXT("dispZ = %f"), dispZ);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("velyZ = %e"), velyZ);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("accelZ = %e"), accelZ);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("state = %d"), state);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

}


// Following must exist for all MAS22 Actors ---------------------------------
void AMAS22_Apocalypse::LogData() {
	int m;
	m = (int(logInterval / dT));
	if (itn % (m + 1) == 0) {
		dataRecords.Add(FDataRecord_Apocal{ time, dispZ, velyZ, accelZ });
		// In above complete the list of your variables to log
	}

}
// Following must exist -------------------------------------------------------
void AMAS22_Apocalypse::ComputeDT() {

	if (bRunning) {
		LogData();
		Computation();
		Visualization(dT);
		BuildHUDLines();
		itn++;
	}
}

void AMAS22_Apocalypse::Computation() { // -------------------------------

	switch (state) {
	case STARTING:
		velyZ = _liftVelocity;
		u[1] = velyZ;
		rightHandSide->setForce(0);
		state = RISING;
		break;

	case RISING:
		if (dispZ >= _releaseHeight) {
			u[1] = 0;
			saveTime = time;
			state = PAUSING;
		}
		break;

	case PAUSING:
		if (time > (saveTime + releaseTime)) {
			forceZ = -_mass * _gravity;
			rightHandSide->setForce(forceZ);
			state = FALLING;
		}
		break;

	case FALLING:
		if (dispZ <= _brakingHeight) {
			forceZ = -_mass * _gravity + _brakingForce;
			rightHandSide->setForce(forceZ);
			state = BRAKING;
		}
		break;

	case BRAKING:
		if (velyZ >= 0.0) {
			rightHandSide->setForce(0);
			u[1] = 0;
			state = STOPPED;
		}
		break;

	case STOPPED:
		break;
	}

	// The following must exist ------------------------------------------
	switch (solver) {
	case SolverEnum::EULER:
		Euler(dT);
		break;
	case SolverEnum::RUNGE_KUTTA_FEHLBERG:
		RKF(dT);
		break;
	case SolverEnum::RUNGE_KUTTA:
		RK(dT);
		break;
	default:
		break;
	}

	//printf("u[0] = %f, u[1] = %f", u[0], u[1]);

	// Grab the state vector values into your model values
	dispZ = u[0];
	velyZ = u[1];
	accelZ = dudt[1];

}

// Numerical Solver Code -------------------------------------------------

void AMAS22_Apocalypse::Euler(double deltaT) { // ------------------------


}

void AMAS22_Apocalypse::RK(double deltaT) { // ---------------------------
	double h = deltaT;
	flag = solverRK->step(u, dudt, time, h, yout, *rightHandSide);
	for (int i = 0; i < NREQUNS; i++) u[1] = yout[i];
	time += deltaT;

}

void AMAS22_Apocalypse::RKF(double deltaT) { // -------------------------

	tout = time + deltaT;

	flag = solverRKF->r8_rkf45(NREQUNS, u, dudt, &time, tout, &relerr, abserr, flag, *rightHandSide);

	if (flag == 3) {
		printErr("Flag = 3 reset");
		flag = 2;
	}
	if (flag == 4) {
		//printErr("Flag = 4 reset");
		flag = 2;
	}
	if (flag == 5) {
		printErr("Flag = 5 exiting");
		exit(0);
	}
	if (flag == 6) {
		printErr("Flag = 6 reset");
		flag = 2;
	}
	if (flag == 7) {
		//printErr("Flag = 7 reset");
		flag = 2;
	}

}

void AMAS22_Apocalypse::Visualization(float deltaTime) { // --------------

	// Move the Car
	FVector newLocation = CarStartLocation;
	newLocation.Z += scaling * dispZ;
	Car->SetRelativeLocation(newLocation);



}

void AMAS22_Apocalypse::UserInput(int key) { // --------------------------

}

// Console commands ------------------------------------------------------

void AMAS22_Apocalypse::SetConsoleAddit(FString name, FString strVal) {

}

void AMAS22_Apocalypse::DumpConsoleAll() { // ----------------------------
	Super::DumpConsoleAll();
}

// Wrap up - dump Octave files -------------------------------------------

void AMAS22_Apocalypse::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	Super::EndPlay(EndPlayReason);

	// Free stepper reserved memory
	if (solver == SolverEnum::RUNGE_KUTTA) solverRK->cleanup();

	FString logFile = FPaths::ProjectDir();
	FString fileName = "OctaveFiles/";
	fileName.Append(actorName);
	fileName.Append(".m");
	logFile.Append(fileName);

	int nrRecords = dataRecords.Num();
	FString stringToSave = "A=[\n";

	for (int i = 0; i < nrRecords; i++) {
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].time));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].dispZ));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].velyZ));
		stringToSave.Append(FString::Printf(TEXT("%e,  \n"), dataRecords[i].accelZ));
	}

	stringToSave.Append("];\n\n");

	// Create names for use in Octave from data matrix A
	stringToSave.Append("time  = A(:,1);\n");
	stringToSave.Append("dispZ  = A(:,2);\n");
	stringToSave.Append("velyZ  = A(:,3);\n");
	stringToSave.Append("accelZ  = A(:,4);\n");

	// Create Octave lines for plotting
	stringToSave.Append("subplot(3,1,1);\n");
	stringToSave.Append("plot(time,dispZ);\n");
	stringToSave.Append("title('dispZ vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('dispZ (m)')\n");
	stringToSave.Append("grid\n");

	stringToSave.Append("subplot(3,1,2);\n");
	stringToSave.Append("plot(time,velyZ);\n");
	stringToSave.Append("title('velyZ vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('velyZ (m)')\n");
	stringToSave.Append("grid\n");

	stringToSave.Append("subplot(3,1,3);\n");
	stringToSave.Append("plot(time,accelZ);\n");
	stringToSave.Append("title('accelZ vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('accelZ (m)')\n");
	stringToSave.Append("grid\n");

	// Write and close file
	FFileHelper::SaveStringToFile(stringToSave, *logFile);

}

void AMAS22_Apocalypse::ApplyConfig() { // -------------------------------
	if (editorOverride == true) {
		FString configFName = FPaths::GeneratedConfigDir().Append(TEXT("MAS22_System.ini"));
		//printf("AMAS22_SkyMaster::ActorInitialize() path to config file = %s", *configFName);
		GConfig->LoadFile(configFName);
		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("scaling"), scaling, configFName);
		printf("AMAS22_Apocalypse::ActorInitialize() >>>>>Config File read: scaling set to %f", scaling);

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("defaultSolver"), defaultSolver, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: solver set to %s", *defaultSolver);
		if (defaultSolver == "RUNGE_KUTTA_FEHLBERG") solver = SolverEnum::RUNGE_KUTTA_FEHLBERG;

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("actorName"), actorName, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: actor Name set to %s", *actorName);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("dT"), dT, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: dT set to %f", dT);

		// Default parameters ----------------------------------------------------------------------

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("mass"), mass, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: mass set to %f", mass);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("gravity"), gravity, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: gravity set to %f", gravity);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("liftVelocity"), liftVelocity, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: liftVelocity set to %f", liftVelocity);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("releaseHeight"), releaseHeight, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: releaseHeight set to %f", releaseHeight);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("brakingHeight"), brakingHeight, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: brakingHeight set to %f", brakingHeight);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("brakingForce"), brakingForce, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: brakingForce set to %f", brakingForce);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("releaseTime"), releaseTime, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: releaseTime set to %f", releaseTime);

		// Set the Initial Conditions -----------------------------------------------------------------------------

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("initDispZ"), initDispZ, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: initDispZ set to %f", initDispZ);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_Apocalypse"), TEXT("initVelyZ"), initVelyZ, configFName);
		printf("MAS22_Apocalypse::ActorInitialize() >>>>>Config File read: initVelyZ set to %f", initVelyZ);


	}
}

