// =======================================================================
//
// MAS22_Apocalypse.h
// Template for MATT
// CBP 03-08-22
// Following code from MLP completed by CBP 
// 11-08-22
// 
// =======================================================================

#pragma once

#include "CoreMinimal.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"
#include "MAS22_Apocalypse.generated.h"

// rhs Class for this actor

class rhs_Apocalypse { // ------------------------------------------------

public:
	// Here comes the parameter list
	double mass;

	// Constructor for this object with parameters
	rhs_Apocalypse(double _mass) : mass(_mass) {

	}

	void operator() (double t, double u[], double dudt[]) {

		accelZ = forceZ / mass;

		dudt[0] = u[1];
		dudt[1] = accelZ;

		//printf(">>>>>> dudt[0] = %f, dudt[1] = %f", dudt[0],dudt[1]);

	}

	void setForce(double _forceZ) {
		forceZ = _forceZ;
		printf(">>>> force asserted = %f <<<<<<<", forceZ);
	}

private:
	double dispZ;
	double velyZ;
	double accelZ;
	double forceZ;
	bool bVelyOverride = true;

}; // --------------------------------------------------------------------



USTRUCT()
struct FDataRecord_Apocal {
	GENERATED_BODY()
		// Actual variables logged depend on specific actor
		double time, dispZ, velyZ, accelZ;
};

// -----------------------------------------------------------------------

UCLASS()
class MAS22_SCIENCEPARK_API AMAS22_Apocalypse : public AMAS22_Actor
{
	GENERATED_BODY()

		/** Capsule collider */
		UPROPERTY(VisibleDefaultsOnly, Category = Mesh)
		class UCapsuleComponent* CapsuleCollider;

	/** Tower mesh */
	UPROPERTY(VisibleDefaultsOnly, Category = Mesh)
		class UStaticMeshComponent* Tower;

	/** Tower top mesh */
	UPROPERTY(VisibleDefaultsOnly, Category = Mesh)
		class UStaticMeshComponent* TowerTop;

	/** Car mesh */
	UPROPERTY(VisibleDefaultsOnly, Category = Mesh)
		class UStaticMeshComponent* Car;

	/////** Destroyed car mesh */
	//UPROPERTY(VisibleDefaultsOnly, Category = Mesh)
		//class UDestructibleComponent* DestroyedCarMesh;

	// Materials ---------------------------------------------------------
	// No materials required
	// UPROPERTY()
	// class UMaterial* baseMaterial;


public:
	AMAS22_Apocalypse();

private:
	// Parameters used in the cpp ----------------------------------------
	double _mass, _gravity, _liftVelocity, _releaseHeight, _brakingHeight, _brakingForce, _releaseTime;
	// Complete above line. Must use underscores

public:
	// Editable variables ------------------------------------------------
	// Complete the following list

	UPROPERTY(EditAnywhere, Config)
		double mass = 1.0;						// CBP 09-08-22
	UPROPERTY(EditAnywhere, Config)
		double gravity = 9.80;
	UPROPERTY(EditAnywhere, Config)
		double liftVelocity = 10;				// CBP 09-08-22
	UPROPERTY(EditAnywhere, Config)
		double releaseHeight = 54;				// CBP 09-08-22
	UPROPERTY(EditAnywhere, Config)
		double brakingHeight = 13.5;			// CBP 09-08-22
	UPROPERTY(EditAnywhere, Config)
		double brakingForce = 39.2;				// CBP 09-08-22
	UPROPERTY(EditAnywhere, Config)
		double releaseTime = 5;					// CBP 09-08-22

	UPROPERTY(EditAnywhere, Config)
		double initDispZ = 0;
	UPROPERTY(EditAnywhere, Config)
		double initVelyZ = 0;


	// -------------------------------------------------------------------

protected:

	virtual void Computation();
	virtual void Visualization(float deltaTime);
	virtual void BuildHUDLines();
	virtual void LogData();
	virtual void UserInput(int key);
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;
	virtual void ApplyConfig() override;

	virtual void Euler(double dT);
	virtual void RKF(double dT);
	virtual void RK(double dT);

	// Overriden functions from MAS22_Actor ---------------------------------

	virtual void ActorInitialize() override;
	virtual void InitializeSolver() override;
	virtual void ComputeDT() override;

	virtual void SetModelParameters() override;
	virtual void UpdateModelParameters() override;
	virtual void GetDefaultParameters() override;

	virtual void SetModelICs() override;
	virtual void UpdateModelICs() override;

	virtual void DumpConsoleAll() override;
	//virtual void InitGlobalLog() override;

	// Console commands specific to this model ------------------------------

	virtual void SetConsoleAddit(FString name, FString strVal) override;

public:

	rhs_Apocalypse* rightHandSide;
	MAS22_Solver_RK* solverRK;
	MAS22_Solver_RKF* solverRKF;

private:
	// Model variables for cpp ---------------------------------------------
	double dispZ;
	double velyZ;
	double accelZ;
	double forceZ;

	int state;

	double saveTime = 0;

	// Addit variables needed by the model ----------------------------------
	FVector CarStartLocation;


	// TArray for data logging ----------------------------------------------
	TArray<FDataRecord_Apocal> dataRecords;

};

