// =======================================================================
//
// MAS22_Meshing.cpp
// CBBP 24-07-22
// Playground to experiment with procedural meshes
// 
// TO DO:
// i) Include Time
// ii) Simplify - try updating lengthmap in ComputeDT() but do not make
//     a completely new mesh
//
// =======================================================================

#include "MAS22_Meshing.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "ProceduralMeshComponent.h"


AMAS22_Meshing::AMAS22_Meshing() {

	PrimaryActorTick.bCanEverTick = false;

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("StaticMesh'/Game/Rolling/beachball.beachball'"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	Ball = CreateDefaultSubobject<UStaticMeshComponent>("Ball");
	Ball->SetStaticMesh(Asset1);
	Ball->SetRelativeLocation(FVector(0, 0, 0));
	Ball->SetRelativeScale3D(FVector(2, 2, 2));
	Ball->SetupAttachment(GetRootComponent());

	pMesh = CreateDefaultSubobject<UProceduralMeshComponent>("pMesh");
	pMesh->SetupAttachment(GetRootComponent());

}

void AMAS22_Meshing::ActorInitialize() {
	Super::ActorInitialize();

	// Set up buffers ----------------------------------------------------

	int nrOfPoints    = (length + 1) * (width + 1);
	int vertexCount   = (width + 1) * (length + 1);
	int triangleCount = width * length * 2 * 3;
	int lengthCount   = vertexCount;
	int normalsCount  = vertexCount;

	vertexArray.Empty();
	vertexArray.AddUninitialized(vertexCount);

	triangleArray.Empty();
	triangleArray.AddUninitialized(triangleCount);

	heightArray.Empty();
	heightArray.AddUninitialized(lengthCount);

	normalsArray.Empty();
	normalsArray.AddUninitialized(normalsCount);

	texCoords.Empty();
	texCoords.AddUninitialized(vertexCount);

	printf("AMAS22_Meshing::ActorInitialize() vertexArray size is %d", vertexArray.Num());
	printf("AMAS22_Meshing::ActorInitialize() triangleArray size is %d", triangleArray.Num());

	// Generate mesh -----------------------------------------------------
	generateMesh();
	//generateHeights();
	//generateGrid(vertexArray, triangleArray, normalsArray, texCoords, FVector2D(Size.X,Size.Y), length, width, heightArray);

	
}

// Travelling Pulse ------------------------------------------------------

void AMAS22_Meshing::generateHeights() {
	int index = 0;
	float val1, val2;
	float b = 0.2;
	float v = 0.1;
	float a = 100;

	for (int x = 0; x < length + 1; x++) {
		for (int y = 0; y < width + 1; y++) {
			//val1 = 2000* b * b * b / (b * b + ((float)x/a - v * time) * ((float)x/a - v * time));
			val1 = 200 * b * b / (b * b + ((float)x / a - v * time) * ((float)x / a - v * time));
			val2 = 200 * b * b / (b * b + ((float)(x-length) / a + v * time) * ((float)(x-length) / a + v * time));
			index = y + (width + 1) * x;
			heightArray[index] = val1 - val2;
		}
	}



}

// Travelling Pulse ------------------------------------------------------

void AMAS22_Meshing::generateHeightsOnePulse() {
	int index = 0;
	float val1;
	float b = 0.2;
	float v = 0.1;
	float a = 100;
	//float maxVal = 0;
	//if (bRunning) {
		for (int x = 0; x < length + 1; x++) {
			for (int y = 0; y < width + 1; y++) {
				//val1 = 2000* b * b * b / (b * b + ((float)x/a - v * time) * ((float)x/a - v * time));
				val1 = 200 * b * b  / (b * b + ((float)x / a - v * time) * ((float)x / a - v * time));
				//if (val1 > maxVal) maxVal = val1;
				index = y + (width + 1) * x;
				heightArray[index] = val1;
			}
		}
		//printf(">>>>>>>>>> maxVal = %f", maxVal);
	//}

}


// Travelling Waves for val1. Standing waves for val1 + val2 -------------
/*
void AMAS22_Meshing::generateHeights() {

	int index = 0;
	float val1, val2;
	float wavesInLength = 4.0F;
	float lambda = (float)(length + 1) / wavesInLength;
	float period = 1.0F;

	printf(">>>>>>>>>>>>> lambda = %f", lambda);

	for (int x = 0; x < length + 1; x++) {
		for (int y = 0; y < width + 1; y++) {
			val1 = 10.0F * FMath::Sin(time*2*M_PI/period  + 2 * M_PI * (float)x / lambda);
			val2 = 10.0F * FMath::Sin(time * 2 * M_PI / period - 2 * M_PI * (float)x / lambda);
			index = y + (width+1) * x;
			heightArray[index] = val1 + val2;
		}
	}
}
*/

void AMAS22_Meshing::generateMesh() {
	pMesh->ClearAllMeshSections();
	generateHeights();
	generateGrid(vertexArray, triangleArray, normalsArray, texCoords, Size, length, width, heightArray);
	const TArray<FLinearColor> EmptyColors{};
	const TArray<FProcMeshTangent> EmptyTangents;
	pMesh->CreateMeshSection_LinearColor(0, vertexArray, triangleArray, normalsArray,texCoords, EmptyColors, TArray<FProcMeshTangent>(), true);
//	pMesh->CreateMeshSection_LinearColor(0, vertexArray, triangleArray, TArray<FVector>(), TArray<FVector2D>(), EmptyColors, TArray<FProcMeshTangent>(), true);
}

void AMAS22_Meshing::ComputeDT() {
	
	//generateGrid(vertexArray, triangleArray, normalsArray, texCoords, FVector2D(Size.X, Size.Y), length, width, heightArray);
	time += dT;
	generateMesh();
	//generateHeights();
	//const TArray<FLinearColor> EmptyColors{};
	//const TArray<FProcMeshTangent> EmptyTangents;
	//pMesh->CreateMeshSection_LinearColor(0, vertexArray, triangleArray, TArray<FVector>(), TArray<FVector2D>(), EmptyColors, TArray<FProcMeshTangent>(), true);

}

void AMAS22_Meshing::generateGrid(TArray<FVector>& InVertices, TArray<int32>& InTriangles, TArray<FVector>& InNormals, TArray<FVector2D>& InTexCoords, const FVector2D InSize, const int32 InLengthSections, const int32 InWidthSections, const TArray<float>& InHeightValues)
{
	// Note the coordinates are a bit weird here since I aligned it to the transform (X is forwards or "up", which Y is to the right)
	// Should really fix this up and use standard X, Y coords then transform into object space?
	const FVector2D SectionSize = FVector2D(InSize.X / InLengthSections, InSize.Y / InWidthSections);
	int32 VertexIndex = 0;
	int32 TriangleIndex = 0;

	const float LengthSectionsAsFloat = static_cast<float>(InLengthSections);
	const float WidthSectionsAsFloat = static_cast<float>(InWidthSections);

	for (int32 X = 0; X < InLengthSections + 1; X++)
	{
		for (int32 Y = 0; Y < InWidthSections + 1; Y++)
		{
			// Create a new vertex
			const int32 NewVertIndex = VertexIndex++;
			const FVector NewVertex = FVector(X * SectionSize.X, Y * SectionSize.Y, InHeightValues[NewVertIndex]);
			InVertices[NewVertIndex] = NewVertex;

			// Note that Unreal UV origin (0,0) is top left
			const float U = static_cast<float>(X) / LengthSectionsAsFloat;
			const float V = static_cast<float>(Y) / WidthSectionsAsFloat;
			InTexCoords[NewVertIndex] = FVector2D(U, V);

			// Once we've created enough verts we can start adding polygons
			if (X > 0 && Y > 0)
			{
				// Each row is InWidthSections+1 number of points.
				// And we have InLength+1 rows
				// Index of current vertex in position is thus: (X * (InWidthSections + 1)) + Y;
				const int32 TopRightIndex = (X * (InWidthSections + 1)) + Y; // Should be same as VertIndex1!
				const int32 TopLeftIndex = TopRightIndex - 1;
				const int32 BottomRightIndex = ((X - 1) * (InWidthSections + 1)) + Y;
				const int32 BottomLeftIndex = BottomRightIndex - 1;

				// Now create two triangles from those four vertices
				// The order of these (clockwise/counter-clockwise) dictates which way the normal will face. 
				InTriangles[TriangleIndex++] = BottomLeftIndex;
				InTriangles[TriangleIndex++] = TopRightIndex;
				InTriangles[TriangleIndex++] = TopLeftIndex;

				InTriangles[TriangleIndex++] = BottomLeftIndex;
				InTriangles[TriangleIndex++] = BottomRightIndex;
				InTriangles[TriangleIndex++] = TopRightIndex;

				// Normals
				const FVector NormalCurrent = FVector::CrossProduct(InVertices[BottomLeftIndex] - InVertices[TopLeftIndex], InVertices[TopLeftIndex] - InVertices[TopRightIndex]).GetSafeNormal();
				InNormals[BottomLeftIndex] = InNormals[BottomRightIndex] = InNormals[TopRightIndex] = InNormals[TopLeftIndex] = NormalCurrent;
			}
		}
	}
}
