// =======================================================================
//
// MAS22_Rolling.cpp
// CBBP 23-07-22
// Playground to experiment with built-in physics
//
// =======================================================================

#include "MAS22_Rolling.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "Components/CapsuleComponent.h"
#include "Components/StaticMeshComponent.h"

AMAS22_Rolling::AMAS22_Rolling() {

	PrimaryActorTick.bCanEverTick = false;
	
	//CapsuleCollider = CreateDefaultSubobject<UCapsuleComponent>("CapsuleCollider");
	//CapsuleCollider->SetupAttachment(GetRootComponent());
	//CapsuleCollider->SetCapsuleHalfHeight(200);
	//CapsuleCollider->SetCapsuleRadius(100);
	//CapsuleCollider->SetRelativeLocation(FVector(0, 0, 0));

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("StaticMesh'/Game/Rolling/beachball.beachball'"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	Ball = CreateDefaultSubobject<UStaticMeshComponent>("Ball");
	Ball->SetStaticMesh(Asset1);
	Ball->SetRelativeLocation(FVector(0, 0, 0));
	Ball->SetRelativeScale3D(FVector(5,5,5));
	Ball->SetupAttachment(GetRootComponent());
	//Ball->SetupAttachment(CapsuleCollider);

	if (Ball == nullptr) {
		printErr("AMAS22_Rolling::AMAS22_Rolling(): Ball is null");
		return;
	} else {
		printf("AMAS22_Rolling::AMAS22_Rolling(): Ball exists");
		printf("AMAS22_Rolling::AMAS22_Rolling(): Pointer = %p", Ball);
	}
}

void AMAS22_Rolling::ActorInitialize() {
	Super::ActorInitialize();

	if (!Ball) {
		printErr("AMAS22_Rolling::ActorInitialize(): Ball is null");
		printErr("AMAS22_Rolling::ActorInitialize(): Pointer = %p", Ball);
		return;
	}

	// Experimenting --------------------------------------------------------

	Ball->SetSimulatePhysics(true);
	Ball->WakeRigidBody();
	Ball->SetLinearDamping(0.0F);
	Ball->SetAngularDamping(0.0F);
	//Ball->SetCollisionObjectType(ECC_WorldStatic);
	
	float Ldamping = Ball->GetLinearDamping();
	printf("Linear Damping = %f", Ldamping);

	if (Ball->IsSimulatingPhysics()) {
		printf("AMAS22_Rolling::ActorInitialize(): Simulating Physics");
	} else {
		printErr("AMAS22_Rolling::ActorInitialize(): Not Simulating Physics");
	}
}

void AMAS22_Rolling::CreatePhysicalMaterial()

{
	float Friction = 0.7f;
	float Restitution = 0.3f;
	float RaiseMassToPower = 0.75f;
	float Density = 1.0f;
	float DestructibleDamageThresholdScale = 1.0f;
	float TireFrictionScale = 1.0f;
	float bOverrideFrictionCombineMode = false;
#if WITH_PHYSX
//	PhysxUserData = FPhysxUserData(this);
#endif
}

void AMAS22_Rolling::ComputeDT() {

}