// =======================================================================
//
// MAS22_Solver_RK.h
// CBP 20-06-22
// Original stepper coding using functors
// Fourth-order Runge-Kutta
// 
// =======================================================================

#pragma once

#include "CoreMinimal.h"

class MAS22_SCIENCEPARK_API MAS22_Solver_RK
{

public:
	MAS22_Solver_RK() {};

	void init(int _nVar) {
		nVar = _nVar;
		yt = (double*)malloc(nVar * sizeof(double));
		dym = (double*)malloc(nVar * sizeof(double));
		dyt = (double*)malloc(nVar * sizeof(double));
	}


	// Classic Runge-Kutta O(4) solver --------------------------------------
	template<class DD>
	int step(double y[], double* dydx, double t, double h, double* yout, DD derivs) {
		double hh = h * 0.5;
		double h6 = h / 6.0;
		double xh = t + hh;

		for (int i = 0; i < nVar; i++) {
			yt[i] = y[i] + hh * dydx[i];
		}
		derivs(xh, yt, dyt);

		for (int i = 0; i < nVar; i++) {
			yt[i] = y[i] + hh * dyt[i];
		}
		derivs(xh, yt, dym);

		for (int i = 0; i < nVar; i++) {
			yt[i] = y[i] + h * dym[i];
			dym[i] += dyt[i];
		}
		derivs(t + h, yt, dyt);
		for (int i = 0; i < nVar; i++) {
			yout[i] = y[i] + h6 * (dydx[i] + dyt[i] + 2.0 * dym[i]);
		}
		return 0;
	}

	void cleanup() {
		free(yt);
		free(dym);
		free(dyt);
	}



private:
	int nVar;
	double* yt;
	double* dym;
	double* dyt;

};


